@@ -27,7 +27,8 @@
 
 #include <krb5.h>
 #include <kadm5/admin.h>
-#include <krb5/adm_proto.h>
+#include <adm_proto.h>
+#include <errno.h>
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
@@ -96,9 +97,12 @@ int locked = 0;
 static void usage()
 {
     fprintf(stderr,
-	 "Usage: %s [-r realm] [-p principal] [-q query] [clnt|local args]\n"
-    "\tclnt args: [-s admin_server[:port]] [[-c ccache]|[-k [-t keytab]]]\n"
-	 "\tlocal args: [-d dbname] [-e \"enc:salt ...\"] [-m]\n", whoami);
+	    "Usage: %s [-r realm] [-p principal] [-q query] [clnt|local args]\n"
+	    "\tclnt args: [-s admin_server[:port]] [[-c ccache]|[-k [-t keytab]]]\n"
+	    "\tlocal args: [-x db_args]* [-d dbname] [-e \"enc:salt ...\"] [-m]\n"
+	    "where,\n\t[-x db_args]* - any number of database specific arguments.\n"
+	    "\t\t\tLook at each database documentation for supported arguments\n",
+	    whoami);
     exit(1);
 }
 
@@ -169,6 +173,21 @@ kadmin_parse_name(name, principal)
     return retval;
 }
 
+static void extended_com_err_fn (const char *myprog, errcode_t code,
+				 const char *fmt, va_list args)
+{
+    if (code) {
+	const char *emsg;
+	emsg = krb5_get_error_message (context, code);
+	fprintf (stderr, "%s: %s ", myprog, emsg);
+	krb5_free_error_message (context, emsg);
+    } else {
+	fprintf (stderr, "%s: ", myprog);
+    }
+    vfprintf (stderr, fmt, args);
+    fprintf (stderr, "\n");
+}
+
 char *kadmin_startup(argc, argv)
     int argc;
     char *argv[];
@@ -183,18 +202,41 @@ char *kadmin_startup(argc, argv)
     krb5_ccache cc;
     krb5_principal princ;
     kadm5_config_params params;
+    char **db_args = NULL;
+    int db_args_size = 0;
+    char *db_name = NULL;
     char *svcname;
 
     memset((char *) &params, 0, sizeof(params));
     
-    retval = krb5_init_context(&context);
+    if (strcmp (whoami, "kadmin.local") == 0)
+	set_com_err_hook(extended_com_err_fn);
+
+    retval = kadm5_init_krb5_context(&context);
     if (retval) {
 	 com_err(whoami, retval, "while initializing krb5 library");
 	 exit(1);
     }
-		     
-    while ((optchar = getopt(argc, argv, "r:p:kq:w:d:s:mc:t:e:O")) != EOF) {
+
+    while ((optchar = getopt(argc, argv, "x:r:p:kq:w:d:s:mc:t:e:ON")) != EOF) {
 	switch (optchar) {
+	case 'x':
+	  db_args_size++;
+	  {
+	    char **temp = realloc( db_args, sizeof(char*) * (db_args_size+1)); /* one for NULL */
+	    if( temp == NULL )
+	      {
+		fprintf(stderr,"%s: Cannot initialize. Not enough memory\n",
+			argv[0]);
+		exit(1);
+	      }
+
+	    db_args = temp;
+	  }
+	  db_args[db_args_size-1] = optarg;
+	  db_args[db_args_size]   = NULL;
+	  break;
+	  
 	case 'r':
 	    def_realm = optarg;
 	    break;
@@ -217,8 +259,33 @@ char *kadmin_startup(argc, argv)
 	    query = optarg;
 	    break;
         case 'd':
-	    params.dbname = optarg;
-	    params.mask |= KADM5_CONFIG_DBNAME;
+	    /* now db_name is not a seperate argument. It has to be passed as part of the db_args */
+	    if( !db_name )
+	    {
+		db_name = malloc( strlen(optarg) + sizeof("dbname="));
+	    } 
+	    else
+	    {
+		db_name = realloc( db_name, strlen(optarg) + sizeof("dbname="));
+	    }
+
+	    strcpy( db_name, "dbname=");
+	    strcat( db_name, optarg );
+
+	    db_args_size++;
+	    {
+		char **temp = realloc( db_args, sizeof(char*) * (db_args_size+1)); /* one for NULL */
+		if( temp == NULL )
+		{
+		    fprintf(stderr,"%s: Cannot initialize. Not enough memory\n",
+			    argv[0]);
+		    exit(1);
+		}
+
+		db_args = temp;
+	    }
+	    db_args[db_args_size-1] = db_name;
+	    db_args[db_args_size]   = NULL;
 	    break;
         case 's':
 	    params.admin_server = optarg;
@@ -244,6 +311,9 @@ char *kadmin_startup(argc, argv)
 	case 'O':
 	    params.mask |= KADM5_CONFIG_OLD_AUTH_GSSAPI;
 	    break;
+	case 'N':
+	    params.mask |= KADM5_CONFIG_AUTH_NOFALLBACK;
+	    break;
 	default:
 	    usage();
 	}
@@ -338,7 +408,7 @@ char *kadmin_startup(argc, argv)
 		if (realm - canon && *(realm - 1) != '\\')
 		    break;
 		else
-		    realm = strchr(realm, '@');
+		    realm = strchr(realm+1, '@');
 	    }
 	    if (realm)
 		*realm++ = '\0';
@@ -347,7 +417,7 @@ char *kadmin_startup(argc, argv)
 		if (cp - canon && *(cp - 1) != '\\')
 		    break;
 		else
-		    cp = strchr(cp, '/');
+		    cp = strchr(cp+1, '/');
 	    }
 	    if (cp != NULL)
 		*cp = '\0';
@@ -414,6 +484,7 @@ char *kadmin_startup(argc, argv)
 					&params,
 					KADM5_STRUCT_VERSION,
 					KADM5_API_VERSION_2,
+					db_args,
 					&handle);
     } else if (use_keytab) {
 	 if (keytab_name)
@@ -427,6 +498,7 @@ char *kadmin_startup(argc, argv)
 				       &params,
 				       KADM5_STRUCT_VERSION,
 				       KADM5_API_VERSION_2,
+				       db_args,
 				       &handle);
     } else {
 	 printf("Authenticating as principal %s with password.\n",
@@ -436,6 +508,7 @@ char *kadmin_startup(argc, argv)
 					   &params,
 					   KADM5_STRUCT_VERSION,
 					   KADM5_API_VERSION_2,
+					   db_args,
 					   &handle);
     }
     if (retval) {
@@ -448,6 +521,12 @@ char *kadmin_startup(argc, argv)
     if (freeprinc)
 	free(princstr);
 
+    if( db_name )
+	free(db_name), db_name=NULL;
+
+    if( db_args )
+	free(db_args), db_args=NULL;
+
     if ((retval = krb5_cc_close(context, cc))) {
 	 com_err(whoami, retval, "while closing ccache %s",
 		 ccache_name);
@@ -585,11 +664,36 @@ void kadmin_cpw(argc, argv)
     krb5_boolean keepold = FALSE;
     krb5_key_salt_tuple *ks_tuple = NULL;
     krb5_principal princ;
+    char **db_args = NULL;
+    int db_args_size = 0;
+
     
     if (argc < 2) {
 	 goto usage;
     }
     for (argv++, argc--; argc > 1; argc--, argv++) {
+	if (!strcmp("-x", *argv)) {
+	    argc--;
+	    if( argc < 1 ) {
+		fprintf( stderr, "change_password: missing db argument\n");
+		goto usage;
+	    }
+	    db_args_size++;
+	    {
+		char **temp = realloc( db_args, sizeof(char*) * (db_args_size+1)); /* one for NULL */
+		if( temp == NULL )
+		{
+		    fprintf(stderr,"change_password: Not enough memory\n");
+		    free( db_args ), db_args = NULL;
+		    exit(1);
+		}
+
+		db_args = temp;
+	    }
+	    db_args[db_args_size-1] = *++argv;
+	    db_args[db_args_size]   = NULL;
+	    continue;
+	}
 	if (!strcmp("-pw", *argv)) {
 	    argc--;
 	    if (argc < 1) {
@@ -625,11 +729,16 @@ void kadmin_cpw(argc, argv)
 	}
 	goto usage;
     }
+    if (*argv == NULL) {
+	com_err("change_password", 0, "missing principal name");
+	goto usage;
+    }
     retval = kadmin_parse_name(*argv, &princ);
     if (retval) {
 	com_err("change_password", retval, "while parsing principal name");
 	if (ks_tuple != NULL)
 	    free(ks_tuple);
+	if( db_args ) free(db_args);
 	return;
     }
     retval = krb5_unparse_name(context, princ, &canon);
@@ -638,6 +747,7 @@ void kadmin_cpw(argc, argv)
 	krb5_free_principal(context, princ);
 	if (ks_tuple != NULL)
 	    free(ks_tuple);
+	if( db_args ) free(db_args);
 	return;
     }
     if (pwarg != NULL) {
@@ -654,10 +764,12 @@ void kadmin_cpw(argc, argv)
 	    com_err("change_password", retval,
 		    "while changing password for \"%s\".", canon);
 	    free(canon);
+	    if( db_args ) free(db_args);
 	    return;
 	}
 	printf("Password for \"%s\" changed.\n", canon);
 	free(canon);
+	if( db_args ) free(db_args);
 	return;
     } else if (randkey) {
 	if (keepold || ks_tuple != NULL) {
@@ -674,10 +786,12 @@ void kadmin_cpw(argc, argv)
 	    com_err("change_password", retval,
 		    "while randomizing key for \"%s\".", canon);
 	    free(canon);
+	    if( db_args ) free(db_args);
 	    return;
 	}
 	printf("Key for \"%s\" randomized.\n", canon);
 	free(canon);
+	if( db_args ) free(db_args);
 	return;
     } else if (argc == 1) {
 	unsigned int i = sizeof (newpw) - 1;
@@ -696,6 +810,7 @@ void kadmin_cpw(argc, argv)
 	    if (ks_tuple != NULL)
 		free(ks_tuple);
 	    krb5_free_principal(context, princ);
+	    if( db_args ) free(db_args);
 	    return;
 	}
 	if (keepold || ks_tuple != NULL) {
@@ -713,15 +828,18 @@ void kadmin_cpw(argc, argv)
 	    com_err("change_password", retval,
 		    "while changing password for \"%s\".", canon);
 	    free(canon);
+	    if( db_args ) free(db_args);
 	    return;
 	}
 	printf("Password for \"%s\" changed.\n", canon);
 	free(canon);
+	if( db_args ) free(db_args);
 	return;
    } else {
 	free(canon);
 	krb5_free_principal(context, princ);
    usage:
+	if( db_args ) free(db_args);
 	if (ks_tuple != NULL)
 	    free(ks_tuple);
 	fprintf(stderr,
@@ -732,6 +850,27 @@ void kadmin_cpw(argc, argv)
    }
 }
 
+static void
+kadmin_free_tl_data( kadm5_principal_ent_t princ )
+{
+    krb5_tl_data *tl_data = princ->tl_data;
+    int n_tl_data         = princ->n_tl_data;
+    int i;
+
+    princ->n_tl_data = 0;
+    princ->tl_data   = NULL;
+
+    for( i = 0; tl_data && (i < n_tl_data); i++ )
+    {
+	krb5_tl_data *next = tl_data->tl_data_next;
+	if( tl_data->tl_data_contents )
+	    free( tl_data->tl_data_contents );
+	free( tl_data );
+	tl_data = next;
+    }
+}
+
+#define KRB5_TL_DB_ARGS 0x7fff
 static int 
 kadmin_parse_princ_args(argc, argv, oprinc, mask, pass, randkey,
 			ks_tuple, n_ks_tuple, caller)
@@ -749,6 +888,7 @@ kadmin_parse_princ_args(argc, argv, oprinc, mask, pass, randkey,
     time_t date;
     time_t now;
     krb5_error_code retval;
+    krb5_tl_data *tl_data, *tail = NULL;
     
     *mask = 0;
     *pass = NULL;
@@ -758,6 +898,42 @@ kadmin_parse_princ_args(argc, argv, oprinc, mask, pass, randkey,
     *randkey = 0;
     for (i = 1; i < argc - 1; i++) {
 	attrib_set = 0;
+	if (strlen(argv[i]) == 2 &&
+	    !strcmp("-x",argv[i])) {
+	    if (++i > argc - 2)
+		return -1;
+
+	    tl_data = malloc( sizeof(krb5_tl_data) );
+	    if( tl_data == NULL )
+	    {
+		fprintf(stderr, "Not enough memory\n");
+		return ENOMEM;
+	    }
+
+	    memset( tl_data, 0, sizeof(krb5_tl_data));
+	    tl_data->tl_data_type = KRB5_TL_DB_ARGS;
+	    tl_data->tl_data_length  = strlen(argv[i])+1;
+	    tl_data->tl_data_contents = strdup(argv[i]);
+
+	    if( tail )
+	    {
+		tail->tl_data_next = tl_data;
+	    }
+	    else
+	    {
+		oprinc->tl_data = tl_data;
+	    }
+	    tail = tl_data;
+	    oprinc->n_tl_data++;
+	    
+	    if( tl_data->tl_data_contents == NULL )
+	    {
+		fprintf(stderr, "Not enough memory\n");
+		return ENOMEM;
+	    }
+	    *mask |= KADM5_TL_DATA;
+	    continue;
+	}
 	if (strlen(argv[i]) == 7 &&
 	    !strcmp("-expire", argv[i])) {
 	    if (++i > argc - 2)
@@ -917,12 +1093,14 @@ kadmin_addprinc_usage(func)
 {
      fprintf(stderr, "usage: %s [options] principal\n", func);
      fprintf(stderr, "\toptions are:\n");
-     fprintf(stderr, "\t\t[-expire expdate] [-pwexpire pwexpdate] [-maxlife maxtixlife]\n\t\t[-kvno kvno] [-policy policy] [-randkey] [-pw password]\n\t\t[-maxrenewlife maxrenewlife]\n\t\t[-e keysaltlist]\n\t\t[{+|-}attribute]\n");
+     fprintf(stderr, "\t\t[-x db_princ_args]* [-expire expdate] [-pwexpire pwexpdate] [-maxlife maxtixlife]\n\t\t[-kvno kvno] [-policy policy] [-randkey] [-pw password]\n\t\t[-maxrenewlife maxrenewlife]\n\t\t[-e keysaltlist]\n\t\t[{+|-}attribute]\n");
      fprintf(stderr, "\tattributes are:\n");
      fprintf(stderr, "%s%s%s",
 	     "\t\tallow_postdated allow_forwardable allow_tgs_req allow_renewable\n",
 	     "\t\tallow_proxiable allow_dup_skey allow_tix requires_preauth\n",
-	     "\t\trequires_hwauth needchange allow_svr password_changing_service\n");
+	     "\t\trequires_hwauth needchange allow_svr password_changing_service\n"
+	     "\nwhere,\n\t[-x db_princ_args]* - any number of database specific arguments.\n"
+	     "\t\t\tLook at each database documentation for supported arguments\n");
 }
 
 static void 
@@ -931,12 +1109,15 @@ kadmin_modprinc_usage(func)
 {
      fprintf(stderr, "usage: %s [options] principal\n", func);
      fprintf(stderr, "\toptions are:\n");
-     fprintf(stderr, "\t\t[-expire expdate] [-pwexpire pwexpdate] [-maxlife maxtixlife]\n\t\t[-kvno kvno] [-policy policy] [-clearpolicy]\n\t\t[-maxrenewlife maxrenewlife] [{+|-}attribute]\n");
+     fprintf(stderr, "\t\t[-x db_princ_args]* [-expire expdate] [-pwexpire pwexpdate] [-maxlife maxtixlife]\n\t\t[-kvno kvno] [-policy policy] [-clearpolicy]\n\t\t[-maxrenewlife maxrenewlife] [{+|-}attribute]\n");
      fprintf(stderr, "\tattributes are:\n");
      fprintf(stderr, "%s%s%s",
 	     "\t\tallow_postdated allow_forwardable allow_tgs_req allow_renewable\n",
 	     "\t\tallow_proxiable allow_dup_skey allow_tix requires_preauth\n",
-	     "\t\trequires_hwauth needchange allow_svr password_changing_service\n");
+	     "\t\trequires_hwauth needchange allow_svr password_changing_service\n"
+	     "\nwhere,\n\t[-x db_princ_args]* - any number of database specific arguments.\n"
+	     "\t\t\tLook at each database documentation for supported arguments\n"
+	     );
 }
 
 void kadmin_addprinc(argc, argv)
@@ -967,7 +1148,8 @@ void kadmin_addprinc(argc, argv)
 				&princ, &mask, &pass, &randkey,
 				&ks_tuple, &n_ks_tuple,
 				"add_principal")) {
-	 kadmin_addprinc_usage("add_principal");
+	 kadmin_addprinc_usage("add_principal"); 
+	 kadmin_free_tl_data( &princ ); /* need to free ks_tuple also??? */
 	 return;
     }
 
@@ -978,6 +1160,7 @@ void kadmin_addprinc(argc, argv)
 	krb5_free_principal(context, princ.principal);
 	if (ks_tuple != NULL)
 	    free(ks_tuple);
+	kadmin_free_tl_data( &princ );
 	return;
     }
 
@@ -1022,6 +1205,7 @@ void kadmin_addprinc(argc, argv)
 		    "while reading password for \"%s\".", canon);
 	    free(canon);
 	    krb5_free_principal(context, princ.principal);
+	    kadmin_free_tl_data( &princ );
 	    return;
 	}
 	pass = newpw;
@@ -1040,6 +1224,7 @@ void kadmin_addprinc(argc, argv)
 	free(canon);
 	if (ks_tuple != NULL)
 	    free(ks_tuple);
+	kadmin_free_tl_data( &princ );
 	return;
     }
     if (randkey) {		/* more special stuff for -randkey */
@@ -1059,6 +1244,7 @@ void kadmin_addprinc(argc, argv)
 	    free(canon);
 	    if (ks_tuple != NULL)
 		free(ks_tuple);
+	    kadmin_free_tl_data( &princ );
 	    return;
 	}
 	princ.attributes &= ~KRB5_KDB_DISALLOW_ALL_TIX;	/* clear notix */
@@ -1071,6 +1257,7 @@ void kadmin_addprinc(argc, argv)
 	    free(canon);
 	    if (ks_tuple != NULL)
 		free(ks_tuple);
+	    kadmin_free_tl_data( &princ );
 	    return;
 	}
     }
@@ -1079,6 +1266,8 @@ void kadmin_addprinc(argc, argv)
     if (ks_tuple != NULL)
 	free(ks_tuple);
     free(canon);
+    kadmin_free_tl_data( &princ );
+
 }
 
 void kadmin_modprinc(argc, argv)
@@ -1134,17 +1323,20 @@ void kadmin_modprinc(argc, argv)
 	free(ks_tuple);
 	kadmin_modprinc_usage("modify_principal");
 	free(canon);
+	kadmin_free_tl_data(&princ);
 	return;
     }
     if (retval) {
 	kadmin_modprinc_usage("modify_principal");
 	free(canon);
+	kadmin_free_tl_data(&princ);
 	return;
     }
     if (randkey) {
 	fprintf(stderr, "modify_principal: -randkey not allowed\n");
 	krb5_free_principal(context, princ.principal);
 	free(canon);
+	kadmin_free_tl_data(&princ);
 	return;
     }
     if (pass) {
@@ -1152,6 +1344,7 @@ void kadmin_modprinc(argc, argv)
 		"modify_principal: -pw not allowed; use change_password\n");
 	krb5_free_principal(context, princ.principal);
 	free(canon);
+	kadmin_free_tl_data(&princ);
 	return;
     }
     retval = kadm5_modify_principal(handle, &princ, mask);
@@ -1160,9 +1353,11 @@ void kadmin_modprinc(argc, argv)
 	com_err("modify_principal", retval,
 		"while modifying \"%s\".", canon);
 	free(canon);
+	kadmin_free_tl_data(&princ);
 	return;
     }
     printf("Principal \"%s\" modified.\n", canon);
+    kadmin_free_tl_data(&princ);
     free(canon);
 }
 
