@@ -40,10 +40,9 @@
 #include	<sys/types.h>
 #include	<gssrpc/rpc.h>
 #include	<krb5.h>
-#include	<k5-int.h>
+#include	<kdb.h>
 #include	<com_err.h>
 #include	<kadm5/kadm_err.h>
-#include	<kadm5/adb_err.h>
 #include	<kadm5/chpass_util_strings.h>
 
 #define KADM5_ADMIN_SERVICE	"kadmin/admin"
@@ -113,7 +112,7 @@ typedef long		kadm5_ret_t;
 #define KADM5_CONFIG_ENCTYPE		0x000200
 #define KADM5_CONFIG_ADBNAME		0x000400
 #define KADM5_CONFIG_ADB_LOCKFILE	0x000800
-#define KADM5_CONFIG_PROFILE		0x001000
+/*#define KADM5_CONFIG_PROFILE		0x001000*/
 #define KADM5_CONFIG_ACL_FILE		0x002000
 #define KADM5_CONFIG_KADMIND_PORT	0x004000
 #define KADM5_CONFIG_ENCTYPES		0x008000
@@ -124,7 +123,6 @@ typedef long		kadm5_ret_t;
 #define KADM5_CONFIG_OLD_AUTH_GSSAPI	0x100000
 #define KADM5_CONFIG_NO_AUTH		0x200000
 #define KADM5_CONFIG_AUTH_NOFALLBACK	0x400000
-
 /*
  * permission bits
  */
@@ -204,18 +202,12 @@ typedef struct _kadm5_policy_ent_t {
 	long		policy_refcnt;
 } kadm5_policy_ent_rec, *kadm5_policy_ent_t;
 
-typedef struct __krb5_key_salt_tuple {
-     krb5_enctype	ks_enctype;
-     krb5_int32		ks_salttype;
-} krb5_key_salt_tuple;
-
 /*
  * Data structure returned by kadm5_get_config_params()
  */
 typedef struct _kadm5_config_params {
      long		mask;
      char *		realm;
-     char *		profile;
      int		kadmind_port;
      int		kpasswd_port;
 
@@ -281,7 +273,7 @@ typedef struct __krb5_realm_params {
 
 #if USE_KADM5_API_VERSION > 1
 krb5_error_code kadm5_get_config_params(krb5_context context,
-					char *kdcprofile, char *kdcenv,
+					int use_kdc_config,
 					kadm5_config_params *params_in,
 					kadm5_config_params *params_out);
 
@@ -304,6 +296,7 @@ kadm5_ret_t    kadm5_init(char *client_name, char *pass,
 #endif
 			  krb5_ui_4 struct_version,
 			  krb5_ui_4 api_version,
+			  char **db_args,
 			  void **server_handle);
 kadm5_ret_t    kadm5_init_with_password(char *client_name,
 					char *pass, 
@@ -315,6 +308,7 @@ kadm5_ret_t    kadm5_init_with_password(char *client_name,
 #endif
 					krb5_ui_4 struct_version,
 					krb5_ui_4 api_version,
+					char **db_args,
 					void **server_handle);
 kadm5_ret_t    kadm5_init_with_skey(char *client_name,
 				    char *keytab,
@@ -326,6 +320,7 @@ kadm5_ret_t    kadm5_init_with_skey(char *client_name,
 #endif
 				    krb5_ui_4 struct_version,
 				    krb5_ui_4 api_version,
+				    char **db_args,
 				    void **server_handle);
 #if USE_KADM5_API_VERSION > 1
 kadm5_ret_t    kadm5_init_with_creds(char *client_name,
@@ -334,6 +329,7 @@ kadm5_ret_t    kadm5_init_with_creds(char *client_name,
 				     kadm5_config_params *params,
 				     krb5_ui_4 struct_version,
 				     krb5_ui_4 api_version,
+				     char **db_args,
 				     void **server_handle);
 #endif
 kadm5_ret_t    kadm5_lock(void *server_handle);
@@ -483,6 +479,8 @@ kadm5_ret_t    kadm5_free_key_data(void *server_handle,
 kadm5_ret_t    kadm5_free_name_list(void *server_handle, char **names, 
 				    int count);
 
+krb5_error_code kadm5_init_krb5_context (krb5_context *);
+
 #if USE_KADM5_API_VERSION == 1
 /*
  * OVSEC_KADM_API_VERSION_1 should be, if possible, compile-time
@@ -594,6 +592,7 @@ ovsec_kadm_ret_t    ovsec_kadm_init(char *client_name, char *pass,
 				    char *service_name, char *realm,
 				    krb5_ui_4 struct_version,
 				    krb5_ui_4 api_version,
+				    char **db_args,
 				    void **server_handle);
 ovsec_kadm_ret_t    ovsec_kadm_init_with_password(char *client_name,
 						  char *pass, 
@@ -601,6 +600,7 @@ ovsec_kadm_ret_t    ovsec_kadm_init_with_password(char *client_name,
 						  char *realm, 
 						  krb5_ui_4 struct_version,
 						  krb5_ui_4 api_version,
+						  char ** db_args,
 						  void **server_handle);
 ovsec_kadm_ret_t    ovsec_kadm_init_with_skey(char *client_name,
 					      char *keytab,
@@ -608,6 +608,7 @@ ovsec_kadm_ret_t    ovsec_kadm_init_with_skey(char *client_name,
 					      char *realm,
 					      krb5_ui_4 struct_version,
 					      krb5_ui_4 api_version,
+					      char **db_args,
 					      void **server_handle);
 ovsec_kadm_ret_t    ovsec_kadm_flush(void *server_handle);
 ovsec_kadm_ret_t    ovsec_kadm_destroy(void *server_handle);
